"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = void 0;
exports.defineProperties = defineProperties;
var _utilEx = require("util-ex");
var _inheritsEx = require("inherits-ex");
var _index = _interopRequireDefault(require("./index.js"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
/**
 * Define properties on the $attributes of the target object/class(prototype).
 *
 * @param {Function|object} aTarget the target class or object
 * @param {PropDescriptors} aProperties the attribute descriptors
 * @param {boolean} [recreate] Whether recreating the $attributes
 * @returns the defined attributes
 */
function defineProperties(aTarget, aProperties, recreate) {
  let nonExported1stChar, vPrototype;
  if ((0, _utilEx.isFunction)(aTarget)) {
    vPrototype = aTarget.prototype;
    nonExported1stChar = vPrototype.nonExported1stChar;
  } else if ((0, _utilEx.isObject)(aTarget)) {
    vPrototype = (0, _inheritsEx.getPrototypeOf)(aTarget);
    nonExported1stChar = aTarget.nonExported1stChar;
  } else {
    throw new TypeError('the target should be a ctor or object!');
  }
  let vAttrs = vPrototype.$attributes;
  const vHasOwnProperty = vPrototype.hasOwnProperty('$attributes');
  const vIsProperties = vAttrs instanceof _index.default;
  if (nonExported1stChar == null) {
    nonExported1stChar = _index.default.prototype.nonExported1stChar;
  }
  if (recreate !== true && !vHasOwnProperty && vIsProperties) {
    vPrototype.$attributes = vAttrs = vAttrs.extends(aProperties, nonExported1stChar);
  } else if (recreate || !vIsProperties) {
    vPrototype.$attributes = vAttrs = (0, _index.default)(aProperties, nonExported1stChar);
  } else {
    if (vAttrs.nonExported1stChar !== nonExported1stChar) {
      vAttrs.nonExported1stChar = nonExported1stChar;
    }
    vAttrs.merge(aProperties);
  }
  return vAttrs;
}
var _default = defineProperties;
exports.default = _default;