"use strict";

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.SimplePropertyManager = SimplePropertyManager;
exports.default = void 0;
var _inheritsEx = require("inherits-ex");
var _utilEx = require("util-ex");
var _abstract = _interopRequireDefault(require("./abstract.js"));
function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }
/* eslint @typescript-eslint/no-invalid-this: 0 */

const getAllOwnKeys = Object.getOwnPropertyNames;
const getOwnPropertyDescriptor = Object.getOwnPropertyDescriptor;
function SimplePropertyManager() {
  if (arguments.length) {
    this.initialize.apply(this, arguments);
  }
}

// merge the methods on the PropertyManager.prototype.
(0, _inheritsEx.extend)(SimplePropertyManager, _abstract.default);

/**
 * Get the all property descriptors of this object except `defaultOptions`
 *
 * @returns {PropDescriptors} the descriptors of properties object
 */
SimplePropertyManager.prototype.getProperties = function getProperties() {
  const result = {};
  const ref = getAllOwnKeys(this);
  for (let i = 0; i < ref.length; i++) {
    const k = ref[i];
    if (k === 'defaultOptions') {
      continue;
    }
    result[k] = getOwnPropertyDescriptor(this, k);
  }
  return result;
};

/**
 * Define properties to this object
 *
 * @param {PropDescriptors} aProperties the descriptors of properties object
 */
SimplePropertyManager.prototype.defineProperties = function defineProperties(aProperties) {
  for (const k in aProperties) {
    let v = aProperties[k];
    if ((0, _utilEx.isArray)(v) || !(0, _utilEx.isObject)(v)) {
      v = {
        value: v
      };
    }
    v.enumerable = v.enumerable !== false;
    const vIsReadonly = v.writable === false || v.get && !v.set;
    if (vIsReadonly && v.exported !== true) {
      v.enumerable = false;
    }
    let value = v.value;
    if (!v.get && !v.set && (0, _utilEx.isObject)(value)) {
      value = (0, _utilEx.cloneObject)(value);
    }
    (0, _inheritsEx.defineProperty)(this, k, value, v);
  }
};
SimplePropertyManager.prototype.assignPropertyTo = function assignPropertyTo(dest, src, name, value, attrs, options) {
  const isExported = options && options.isExported;
  if (!attrs) {
    attrs = this.getProperties();
  }
  const vAttr = attrs[name];
  if (vAttr && vAttr.enumerable) {
    const vCanAssign = !isExported && (vAttr.writable || vAttr.set) || isExported && value !== undefined && name[0] !== this.nonExported1stChar;
    if (vCanAssign) {
      if (isExported && value != null) {
        if ((0, _utilEx.isFunction)(value.toObject)) {
          value = value.toObject(options);
        } else if ((0, _utilEx.isFunction)(value.toJSON)) {
          value = value.toJSON();
        }
      }
      dest[name] = value;
    }
  }
};
var _default = SimplePropertyManager;
exports.default = _default;